/**
 * Created by IntelliJ IDEA.
 * User: Kilho Shin
 * Date: 11/02/21
 * Time: 12:00
 * To change this template use File | Settings | File Templates.

  This file is part of Mapping Kernels for Trees.

  Mapping Kernels for Trees is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mapping Kernels for Trees is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mapping Kernels for Trees.  If not, see <http://www.gnu.org/licenses/>.

  For more details about this code, please look at the companion paper
  "Mapping Kernels for Trees", K. Shin, M.Cuturi, T.Kuboyama, Proceedings of
  the International Conference on Machine Learning 2011.

 */

import scala.actors._
import scala.actors.Actor._
import scala.collection.mutable.Map
import scala.util.matching.Regex

class TParser(
  val src: List[(Int, String)],
  val refs: Map[Int, List[RefNode]],
  val db: Map[Int, Option[Vertex]],
  val cubic: Boolean) {
  val pttn = new Regex("""([^\(\)\s]+)?\s*([\(\)])?\s*(\S.*)?""")

  def parse {
    src.foreach(x=>(parse_element(x._1,x._2)))
  }

  private def parse_element(index: Int, s: String) {
    var rest = s
    val virtualRoot = new Vertex("VIRTUAL")
    var crrVertex = virtualRoot
    var newVertex = virtualRoot
    var err = false
    while (rest != null && err == false) {
      val pttn(label, delim, temp) = rest
      rest = temp
      if(label != null) {
        newVertex = new Vertex(crrVertex, label, crrVertex.depth+1)
        crrVertex.addChild(newVertex)
        delim match{
          case "(" =>
            crrVertex = newVertex
          case ")" =>
            if(crrVertex == virtualRoot)
              err = true
            else {
              finish_new_vertex
              finish_current_vertex
              crrVertex = crrVertex.parent
            }
          case _ =>
            finish_new_vertex
        }
      } else {
        delim match{
          case "(" =>
              err = true
          case ")" =>
            if(crrVertex == virtualRoot)
              err = true
            else {
              finish_current_vertex
              crrVertex = crrVertex.parent
            }
          case _ =>
        }
      }
    }
    if(err == true || crrVertex != virtualRoot)
      db += index->None
    else
      db += index->Some(virtualRoot)


    def finish_current_vertex {
      crrVertex.height = crrVertex.children.foldLeft(0)(_ max _.height) + 1
      crrVertex.size = crrVertex.children.foldLeft(0)(_+_.size)+1
      crrVertex.refAsTree = register_if_not_registered(null, List(crrVertex))
      crrVertex.refOfChildren =
        register_if_not_registered(crrVertex.children.head.refAsTree, crrVertex.children.tail)
      register_index_and_CAB_if_required(crrVertex.refAsTree)
    }

    def finish_new_vertex {
      newVertex.height = 1
      newVertex.size = 1
      newVertex.refAsTree = register_if_not_registered(null, List(newVertex))
      register_index_and_CAB_if_required(newVertex.refAsTree)
    }

    def register_if_not_registered(ref: RefNode, vertices: List[Vertex]): RefNode = {
      var anchor: RefNode = ref
      var list: List[Vertex] = vertices
      if(list != Nil) {
        if(anchor == null) {
          val top = list.head
          list = list.tail
          top.refAsTree match {
            case null =>
              if(refs.isDefinedAt(top.height)){
                refs(top.height).find(_.entity isIsomTo top) match {
                  case Some(x) =>
                    anchor = x
                  case None =>
                    anchor = new RefNode(top)
                    anchor.size = top.size
                    refs(top.height) ::= anchor
                }
              } else {
                anchor = new RefNode(top)
                anchor.size = top.size
                refs += top.height -> List(anchor)
              }
            case x: RefNode =>
              anchor = x
          }
        }
        for(top <- list) {
          val found: Option[RefNode] = top.refAsTree match {
            case null =>
              anchor.next.find(_.entity isIsomTo top)
            case x: RefNode =>
              anchor.next.find(_.entity.refAsTree == x)
          }
          found match {
            case Some(x) =>
              anchor = x
            case None =>
              val newRev = new RefNode(top)
              newRev.size = anchor.size + top.size
              newRev.prev = anchor
              anchor.next ::= newRev
              anchor = newRev
          }
        }
      }
      anchor
    }

    /*
    def find_longest_match(ref: RefNode, vertices: List[Vertex]): (RefNode, List[Vertex]) = {
      var anchor: RefNode = ref
      var list: List[Vertex] = vertices
      var matching: Boolean = true
      if(list != Nil) {
        if(anchor == null) {
          val top = list.head
          top.refAsTree match {
            case null =>
              if(!refs.isDefinedAt(top.height))
                matching = false
              else {
                refs(top.height).find(_.entity isIsomTo top) match {
                  case Some(x) =>
                    anchor = x
                    list = list.tail
                  case None =>
                    matching = false
                }
              }
            case x: RefNode =>
              anchor = x
              list = list.tail
          }
          while(list != Nil && matching) {
            val top = list.head
            list = list.tail
            val found: Option[RefNode] = top.refAsTree match {
              case null =>
                anchor.next.find(_.entity isIsomTo top)
              case x: RefNode =>
                anchor.next.find(_.entity.refAsTree == x)
            }
            found match {
              case Some(x) =>
                anchor = x
              case None =>
                matching = false
            }
          }
        }
      }
      (anchor, list)
    }
    */

    def register_index_and_CAB_if_required(ref: RefNode) {
      if(ref != null && !ref.referencedFrom.contains(index)) {
        ref.referencedFrom += index
        if(cubic) {
          var indicator: RefNode = null
          if(ref.prev == null)
            indicator = ref.entity.refOfChildren
          else
            indicator = register_if_not_registered(ref.prev, ref.entity.children)
          ref.cut_and_branch = indicator
          while(indicator != null && !indicator.referencedFrom.contains(index)) {
            register_index_and_CAB_if_required(indicator)
            indicator = indicator.prev
          }
        } else if(ref.prev == null) {
          var indicator: RefNode = ref.entity.refOfChildren
          while(indicator != null && !indicator.referencedFrom.contains(index)) {
            register_index_and_CAB_if_required(indicator)
            indicator = indicator.prev
          }
        }
      }
    }
/*

    def findAnchor(ref: RefNode, list: List[Vertex]): (RefNode, List[Vertex]) = {
      if(list != Nil)
        ref.next.find(_.entity.refAsTree == list.head.refAsTree) match{
          case Some(r: RefNode) => findAnchor(r, list.tail)
          case None => (ref, list)
        }
      else
        (ref, list)
    }

    def find_maximum_match(ref: RefNode, list: List[Vertex]): (RefNode, List[Vertex]) = {
      val top = list.head.refAsTree
      (ref, list) match {
        case (_, Nil) => (ref, list)
        case (null, _) =>
          if(!refs.isDefinedAt(top.height))
            (ref, list)
          else {
            refs(top.height).find(_.entity.refAsTree == top) match {
              case Some(x: RefNode) => find_maximum_match(x, list.tail)
              case None => (ref, list)
            }
          }
        case _ =>
          ref.next.find(_.entity.refAsTree == top) match {
            case Some(x: RefNode) => find_maximum_match(x, list.tail)
            case None => (ref, list)
          }
      }
    }


    def finishCrrVertex = {
      val h = crrVertex.children.foldLeft(0)(_ max _.height) + 1
      crrVertex.height = h
      crrVertex.size = crrVertex.children.foldLeft(0)(_+_.size)+1
      find_longest_match(null, List(crrVertex)) match {
        case r: RefNode =>
          crrVertex.refAsTree = r
          crrVertex.refOfChildren = r.entity
      }
      if(refs isDefinedAt h)
        refs(h).find(_.entity isIsomTo crrVertex) match{
          case Some(r) =>
            crrVertex.refAsTree = r
            crrVertex.refOfChildren = r.entity.refOfChildren //不要？
          case None => rgstCrrVertex
        }
      else
        rgstCrrVertex
      mg ! AddIndex(i, crrVertex.refAsTree, crrVertex.refOfChildren)
    }


    def rgstCrrVertex = {
      val ref = register_if_not_registered(null, List(crrVertex))
      crrVertex.refAsTree = ref
      /*
      mg !? RgstTree(crrVertex) match{
        case r: RefNode =>
          crrVertex.refAsTree = r
        case _ => //Error
      }
      */
      crrVertex.refOfChildren =
        register_if_not_registered(crrVertex.children.head.refAsTree, crrVertex.children.tail)
      /*
      val ref = crrVertex.children.head.refAsTree
      val tpl = find_longest_match(ref, crrVertex.children.tail)
      if(tpl._2 != Nil)
        mg !? RgstFrst(tpl) match{
          case r: RefNode =>
            crrVertex.refOfChildren = r
          case _ => //Error
        }
      else
        crrVertex.refOfChildren = tpl._1
      */
      if(cubic)
        ref.cut_and_branch = register_cut_and_branches(ref)
    }

    def finishNewVertex = {
      if(refs isDefinedAt 1)
        refs(1).find(_.entity isIsomTo newVertex) match{
          case Some(r) => newVertex.refAsTree = r
          case None => rgstNewVertex
        }
      else
        rgstNewVertex
      mg ! AddIndex(i, newVertex.refAsTree, null)
    }

    def rgstNewVertex = {
      mg !? RgstTree(newVertex) match{
      case r: RefNode =>
        newVertex.refAsTree = r
      case _ =>
      }
    }



    def sendRgstTree(v: Vertex) = {
      mg !? RgstTree(v) match{
        case r: RefNode => v.refAsTree = r
        case _ =>
      }
    }
    */

  }
}

